@@ -11,14 +11,12 @@ from __future__ import print_function
 # mapping ::= MAPPING-START (node node)* MAPPING-END
 
 import sys
-
-
-from ruamel.yaml.error import YAMLError
+from ruamel.yaml.error import YAMLError, YAMLStreamError
 from ruamel.yaml.events import *                                     # NOQA
 from ruamel.yaml.compat import utf8, text_type, PY2, nprint, dbg, DBG_EVENT, \
     check_anchorname_char
 
-if sys.version_info >= (3, 5, 2):
+if False:  # MYPY
     from typing import Any, Dict, List, Union, Text  # NOQA
     from ruamel.yaml.compat import StreamType  # NOQA
 
@@ -58,13 +56,13 @@ class Emitter(object):
                  top_level_colon_align=None, prefix_colon=None, dumper=None):
         # type: (StreamType, Any, int, int, bool, Any, int, bool, Any, Any) -> None
         self.dumper = dumper
-        if self.dumper is not None:
+        if self.dumper is not None and getattr(self.dumper, '_emitter', None) is None:
             self.dumper._emitter = self
-        # The stream should have the methods `write` and possibly `flush`.
         self.stream = stream
 
         # Encoding can be overriden by STREAM-START.
         self.encoding = None  # type: Union[None, Text]
+        self.allow_space_break = None
 
         # Emitter is a state machine with a stack of states to handle nested
         # structures.
@@ -109,17 +107,20 @@ class Emitter(object):
         # Formatting details.
         self.canonical = canonical
         self.allow_unicode = allow_unicode
-        self.block_seq_indent = block_seq_indent if block_seq_indent else 0  # type: ignore
+        # set to False to get "\Uxxxxxxxx" for non-basic unicode like emojis
+        self.unicode_supplementary = sys.maxunicode > 0xffff
+        self.block_seq_indent = block_seq_indent if block_seq_indent else 0
         self.top_level_colon_align = top_level_colon_align
         self.best_indent = 2
-        if indent and 1 < indent < 10:  # type: ignore
+        self.requested_indent = indent  # specific for literal zero indent
+        if indent and 1 < indent < 10:
             self.best_indent = indent
         # if self.best_indent < self.block_seq_indent + 1:
         #     self.best_indent = self.block_seq_indent + 1
         self.best_width = 80
-        if width and width > self.best_indent*2:  # type: ignore
+        if width and width > self.best_indent*2:
             self.best_width = width
-        self.best_line_break = u'\n'
+        self.best_line_break = u'\n'  # type: Any
         if line_break in [u'\r', u'\n', u'\r\n']:
             self.best_line_break = line_break
 
@@ -134,6 +135,23 @@ class Emitter(object):
         self.analysis = None  # type: Any
         self.style = None  # type: Any
 
+    @property
+    def stream(self):
+        # type: () -> Any
+        try:
+            return self._stream
+        except AttributeError:
+            raise YAMLStreamError('output stream needs to specified')
+
+    @stream.setter
+    def stream(self, val):
+        # type: (Any) -> None
+        if val is None:
+            return
+        if not hasattr(val, 'write'):
+            raise YAMLStreamError('stream argument needs to have a write() method')
+        self._stream = val
+
     def dispose(self):
         # type: () -> None
         # Reset the state attributes (to clear self-references)
@@ -615,6 +633,9 @@ class Emitter(object):
             self.prepared_tag = self.prepare_tag(tag)
         if self.prepared_tag:
             self.write_indicator(self.prepared_tag, True)
+            if self.sequence_context and not self.flow_level and \
+               isinstance(self.event, ScalarEvent):
+                self.no_newline = True
         self.prepared_tag = None
 
     def choose_scalar_style(self):
@@ -624,12 +645,13 @@ class Emitter(object):
         if self.event.style == '"' or self.canonical:
             return '"'
         if (not self.event.style or self.event.style == '?') and \
-           self.event.implicit[0]:
+           (self.event.implicit[0] or not self.event.implicit[2]):
             if (not (self.simple_key_context and
                      (self.analysis.empty or self.analysis.multiline)) and
                 (self.flow_level and self.analysis.allow_flow_plain or
                     (not self.flow_level and self.analysis.allow_block_plain))):
                 return ''
+        self.analysis.allow_block = True
         if self.event.style and self.event.style in '|>':
             if (not self.flow_level and not self.simple_key_context and
                     self.analysis.allow_block):
@@ -841,7 +863,9 @@ class Emitter(object):
                 line_breaks = True
             if not (ch == u'\n' or u'\x20' <= ch <= u'\x7E'):
                 if (ch == u'\x85' or u'\xA0' <= ch <= u'\uD7FF' or
-                        u'\uE000' <= ch <= u'\uFFFD') and ch != u'\uFEFF':
+                        u'\uE000' <= ch <= u'\uFFFD' or
+                        (self.unicode_supplementary and
+                         (u'\U00010000' <= ch <= u'\U0010FFFF'))) and ch != u'\uFEFF':
                     # unicode_characters = True
                     if not self.allow_unicode:
                         special_characters = True
@@ -900,9 +924,12 @@ class Emitter(object):
 
         # Spaces followed by breaks, as well as special character are only
         # allowed for double quoted scalars.
-        if space_break or special_characters:
-            allow_flow_plain = allow_block_plain = \
-                allow_single_quoted = allow_block = False
+        if special_characters:
+            allow_flow_plain = allow_block_plain = allow_single_quoted = allow_block = False
+        elif space_break:
+            allow_flow_plain = allow_block_plain = allow_single_quoted = False
+            if not self.allow_space_break:
+                allow_block = False
 
         # Although the plain scalar writer supports breaks, we never emit
         # multiline plain scalars.
@@ -935,7 +962,7 @@ class Emitter(object):
     def write_stream_start(self):
         # type: () -> None
         # Write BOM if needed.
-        if self.encoding and self.encoding.startswith('utf-16'):  # type: ignore
+        if self.encoding and self.encoding.startswith('utf-16'):
             self.stream.write(u'\uFEFF'.encode(self.encoding))
 
     def write_stream_end(self):
@@ -970,9 +997,9 @@ class Emitter(object):
             self.whitespace = True
             data = u' '*(indent-self.column)
             self.column = indent
-            if self.encoding:  # type: ignore
+            if self.encoding:
                 data = data.encode(self.encoding)
-            self.stream.write(data)  # type: ignore
+            self.stream.write(data)
 
     def write_line_break(self, data=None):
         # type: (Any) -> None
@@ -989,23 +1016,28 @@ class Emitter(object):
     def write_version_directive(self, version_text):
         # type: (Any) -> None
         data = u'%%YAML %s' % version_text
-        if self.encoding:  # type: ignore
+        if self.encoding:
             data = data.encode(self.encoding)
-        self.stream.write(data)  # type: ignore
+        self.stream.write(data)
         self.write_line_break()
 
     def write_tag_directive(self, handle_text, prefix_text):
         # type: (Any, Any) -> None
         data = u'%%TAG %s %s' % (handle_text, prefix_text)
-        if self.encoding:  # type: ignore
+        if self.encoding:
             data = data.encode(self.encoding)
-        self.stream.write(data)  # type: ignore
+        self.stream.write(data)
         self.write_line_break()
 
     # Scalar streams.
 
     def write_single_quoted(self, text, split=True):
         # type: (Any, Any) -> None
+        if self.root_context:
+            if self.requested_indent is not None:
+                self.write_line_break()
+                if self.requested_indent != 0:
+                    self.write_indent()
         self.write_indicator(u'\'', True)
         spaces = False
         breaks = False
@@ -1079,6 +1111,11 @@ class Emitter(object):
 
     def write_double_quoted(self, text, split=True):
         # type: (Any, Any) -> None
+        if self.root_context:
+            if self.requested_indent is not None:
+                self.write_line_break()
+                if self.requested_indent != 0:
+                    self.write_indent()
         self.write_indicator(u'"', True)
         start = end = 0
         while end <= len(text):
@@ -1218,7 +1255,7 @@ class Emitter(object):
                             self.write_line_break()
                         else:
                             self.write_line_break(br)
-                    if ch is not None:
+                    if ch is not None and (not self.root_context or self.requested_indent):
                         self.write_indent()
                     start = end
             else:
@@ -1237,15 +1274,20 @@ class Emitter(object):
     def write_plain(self, text, split=True):
         # type: (Any, Any) -> None
         if self.root_context:
-            self.open_ended = True
+            if self.requested_indent is not None:
+                self.write_line_break()
+                if self.requested_indent != 0:
+                    self.write_indent()
+            else:
+                self.open_ended = True
         if not text:
             return
         if not self.whitespace:
             data = u' '
             self.column += len(data)
-            if self.encoding:  # type: ignore
+            if self.encoding:
                 data = data.encode(self.encoding)
-            self.stream.write(data)  # type: ignore
+            self.stream.write(data)
         self.whitespace = False
         self.indention = False
         spaces = False
@@ -1265,9 +1307,9 @@ class Emitter(object):
                     else:
                         data = text[start:end]
                         self.column += len(data)
-                        if self.encoding:  # type: ignore
+                        if self.encoding:
                             data = data.encode(self.encoding)
-                        self.stream.write(data)  # type: ignore
+                        self.stream.write(data)
                     start = end
             elif breaks:
                 if ch not in u'\n\x85\u2028\u2029':
@@ -1286,9 +1328,9 @@ class Emitter(object):
                 if ch is None or ch in u' \n\x85\u2028\u2029':
                     data = text[start:end]
                     self.column += len(data)
-                    if self.encoding:  # type: ignore
+                    if self.encoding:
                         data = data.encode(self.encoding)
-                    self.stream.write(data)  # type: ignore
+                    self.stream.write(data)
                     start = end
             if ch is not None:
                 spaces = (ch == u' ')
